Rem  Access.Bas
Rem
Rem  Check Access Conditions

Declare Sub CheckCompactSecAttr (INSMask@, SA$)
Declare Function CheckSCByte (SCByte@)
Declare Sub CheckExpandedSecAttr (INSMask@, SA$)
Declare Function InCreationState (filenum)
Declare Function CheckSCDataObject (Tag@, ByVal S$)
Declare Function FindSecurityEnvironment$ (SEID@)
Declare Function CheckSecureMessaging (SE$)
Declare Function CheckExtAuth (SE$)
Declare Function CheckUserAuth (SE$)
Declare Function CheckAuthCRT (SE$, KeyRef@, UQMask)
Declare Function InvalidatedPassword (key@)

Rem  Sub CheckAccess (EFINSMask@, DFINSMask@)
Rem
Rem  Check the Access Conditions for the given instruction masks

Sub CheckAccess (EFINSMask@, DFINSMask@)

  Private Value$

  If EFINSMask@ <> 0 And Not InCreationState (EFFilenum) Then

    Rem  Check EF access conditions

    Call TLVReadValue (EFFilenum, Chr$(TagFCP)+Chr$(TagCompactSecAttr), Value$)
    Call CheckCompactSecAttr (EFINSMask@, Value$)

    Call TLVReadValue (EFFilenum, Chr$(TagFCP)+Chr$(TagExpandedSecAttr), Value$)
    Call CheckExpandedSecAttr (EFINSMask@, Value$)

  End If

  Rem  Check DF access conditions (even if DFINSMask@ is zero)

  Private filenum : filenum = Open "FCP" For Binary Access Read Shared
  If FileError <> 0 Then SW1SW2 = swMissingSMObject : Exit

  If Not InCreationState (filenum) Then
    Call TLVReadValue (filenum, Chr$(TagFCP)+Chr$(TagCompactSecAttr), Value$)
    Call CheckCompactSecAttr (DFINSMask@, Value$)

    Call TLVReadValue (filenum, Chr$(TagFCP)+Chr$(TagExpandedSecAttr), Value$)
    Call CheckExpandedSecAttr (DFINSMask@, Value$)
  End If

  Close #filenum

End Sub

Rem  Sub CheckCompactSecAttr (INSMask@, SA$)
Rem
Rem  ISO/IEC 7816-4 section 5.3.3.4.1: Compact format

Sub CheckCompactSecAttr (INSMask@, SA$)

  While Len(SA$) <> 0

    Private SCFound = False ' Relevant Security Condition not yet found
    Private AccessMode@ : AccessMode@ = GetFirstByte@ (SA$)

    Private Mask@ : Mask@ = &H40

    Private I : For I = 1 To 7

      If AccessMode@ And Mask@ Then ' Security Condition is present

        Private SecurityCondition@
        SecurityCondition@ = GetFirstByte@ (SA$)

        If INSMask@ = Mask@ Then

          SCFound = True
          If CheckSCByte (SecurityCondition@) Then Exit Sub ' Satisfied

        End If ' If INSMask@ = Mask@

      End If ' If AccessMode@ And Mask@

      Mask@ = Mask@ / 2

    Next I

  Wend ' While Len(SA$) <> 0

  Rem  If there was at least one relevant Security Condition, but none were
  Rem  satisfied, then fail:
  If SCFound Then SW1SW2 = swSecurityStatus : Exit

End Sub

Rem  Function CheckSCByte (SCByte@)
Rem
Rem  Table 21 - Security condition byte

Function CheckSCByte (SCByte@)

  Select Case SCByte@

    Case &H00 : CheckSCByte = True : Exit Function

    Case &HFF : Exit Function ' Return False

    Case Else

      Private SE$
      SE$ = FindSecurityEnvironment$ (SCByte@ And &H0F)
  
      If SCByte@ And &H40 Then ' Secure messaging
        If CheckSecureMessaging (SE$) Then
          If (SCByte@ And &H80) = 0 Then_ ' At least one condition
            CheckSCByte = True : Exit Function
        Else
          If SCByte@ And &H80 Then_ ' All conditions
            Exit Function ' Return False
        End If
      End If
  
      If SCByte@ And &H20 Then ' External authentication
        If CheckExtAuth (SE$) Then
          If (SCByte@ And &H80) = 0 Then_ ' At least one condition
            CheckSCByte = True : Exit Function
        Else
          If SCByte@ And &H80 Then_ ' All conditions
            Exit Function ' Return False
        End If
      End If
    
      If SCByte@ And &H10 Then ' User authentication
        If CheckUserAuth (SE$) Then
          If (SCByte@ And &H80) = 0 Then_ ' At least one condition
            CheckSCByte = True : Exit Function
        Else
          If SCByte@ And &H80 Then_ ' All conditions
            Exit Function ' Return False
        End If
      End If

  End Select

  Rem  Return False

End Function

Rem  Sub CheckExpandedSecAttr (INSMask@, SA$)
Rem
Rem  ISO/IEC 7816-4 section 5.3.3.4.2: Expanded format

Sub CheckExpandedSecAttr (INSMask@, SA$)

  While Len(SA$) <> 0

    Private Tag@ : Tag@ = GetFirstByte@ (SA$)
    Private Len@ : Len@ = GetFirstByte@ (SA$)
    If Len@ > Len(SA$) Then SW1SW2 = swBadSMObject : Exit
    Private Relevant : Relevant = False

    If Tag@ = TagAccessModeByte Then
      If Len@ <> 1 Then SW1SW2 = swBadSMObject : Exit
      Tag@ = GetFirstByte@ (SA$) ' Access Mode Byte
      If (INSMask@ And Tag@) <> 0 Then Relevant = True
    ElseIf Tag@ >= &H81 And Tag@ <= &H8F Then
      Do
        Private Match : Match = True
        If (Tag@ And &H08) <> 0 Then
          Len@ = Len@ - 1
          If CLA <> GetFirstByte@ (SA$) Then Match = False
        End If
        If (Tag@ And &H04) <> 0 Then
          Len@ = Len@ - 1
          If (INS And &HFE) <> GetFirstByte@ (SA$) Then Match = False
        End If
        If (Tag@ And &H02) <> 0 Then
          Len@ = Len@ - 1
          If P1 <> GetFirstByte@ (SA$) Then Match = False
        End If
        If (Tag@ And &H01) <> 0 Then
          Len@ = Len@ - 1
          If P2 <> GetFirstByte@ (SA$) Then Match = False
        End If
        If Match Then Relevant = True
      Loop While Len@ <> 0
    Else
      SW1SW2 = swMissingSMObject : Exit
    End If

    If Relevant And Not CheckSCDataObject (TagSCAnd, Left$(SA$, Len@)) Then_
      SW1SW2 = swSecurityStatus : Exit

    SA$ = Mid$(SA$, Len@+1)

  Wend

End Sub

Rem  Function FindSecurityEnvironment$ (SEID@)
Rem
Rem  Look for the given Security Environment ID in the FCP of successive
Rem  parent DF's

Function FindSecurityEnvironment$ (SEID@)
  If SEID@ = 0 Or SEID@ = 15 Then SW1SW2 = swBadSMObject : Exit

  Private dir$ : dir$ = CurrentDF
  If EFFilenum Then Close #EFFilenum

  Do
    Rem  Open the file "FCP" in directory dir$
    Private fnFCP : fnFCP = Open dir$+"\FCP" For Binary Access Read Shared
    If FileError = 0 Then
      Rem  Look for the Security Environment EF
      Private FileID$
      Call TLVReadValue (fnFCP, Chr$(TagFCP)+Chr$(TagSEEF), FileID$)
      Close #fnFCP

      If Len(FileID$) <> 0 Then

        If Len(FileID$) <> 2 Then SW1SW2 = swBadSMObject : Exit

        Rem  Open the Security Environment EF
        Private FileID2 As String*2 : FileID2 = FileID$
        Private FileID At FileID2
        Call OpenEF (FileID, EFTypeRecord)

        Rem  According to ISO, SE's should be encapsulated in a '7B'
        Rem  template. But SCOSTA ignores this:
        Private recnum : For recnum = 1 To NumRecords()
          Private S$ : S$ = Record$ (recnum)
          If Asc(S$) <> &H80 Or Asc(S$(2)) <> 1 Then_
            SW1SW2 = swFileStructure : Exit
          If Asc(S$(3)) = SEID@ Then
            Close #EFFilenum
            FindSecurityEnvironment$ = Mid$(S$, 4)
            GoTo RestoreCurrentEF
          End If
        Next recnum
        Close #EFFilenum

      End If ' If Len(FileID$) <> 0

    End If ' If FileError = 0

    Close #fnFCP ' We may already have closed it, but never mind
    FileError = 0

    Rem  Parent directory
    If Len(dir$) < 8 Then SW1SW2 = swMissingSMObject : Exit
    dir$ = Left$ (dir$, Len(dir$)-5)

  Loop

RestoreCurrentEF:

  EFFilenum = 0
  If CurrentEF <> &HFFFF Then Call OpenEF(CurrentEF, EFTypeAny)

End Function

Function CheckSecureMessaging (SE$)
  SW1SW2 = swFunctionNotSupported : Exit
End Function

Rem  Function CheckExtAuth (SE$)
Rem
Rem  Check that SecStat.ExtAuthRef satisfies the ExtAuth field in SE$

Function CheckExtAuth (SE$)

  CheckExtAuth = CheckAuthCRT (SE$, SecStat.ExtAuthRef, UQExtAuth)

End Function

Rem  Function CheckUserAuth (SE$)
Rem
Rem  Check that SecStat.UserAuthRef satisfies the ExtAuth field in SE$

Function CheckUserAuth (SE$)

  CheckUserAuth = CheckAuthCRT (SE$, SecStat.UserAuthRef, UQUserAuthKB)

End Function

Rem  Function CheckAuthCRT (SE$, KeyRef@, UQMask)
Rem
Rem  Common code for CheckExtAuth and CheckUserAuth
Rem
Rem  The CRT is assumed to have the form
Rem
Rem     '83'  Secret key (may be repeated)
Rem     '95'  Usage qualifier byte

Function CheckAuthCRT (SE$, KeyRef@, UQMask)

  Private CRT$ : CRT$ = TLVScanString$ (TagCrtAT, SE$)
  Private Tag@ : Tag@ = GetFirstByte@ (CRT$)

  Rem  Look for the key  
  Do
    If Tag@ <> TagCRTSecretKey Or GetFirstByte@ (CRT$) <> 1 Then_
      SW1SW2 = swMissingSMObject : Exit

    Private Key@ : Key@ = GetFirstByte@ (CRT$)
    If Key@ = KeyRef@ Or_
      (UQMask = UQUserAuthKB And InvalidatedPassword (Key@)) Then_
        CheckAuthCRT = True

    Tag@ = GetFirstByte@ (CRT$)
  Loop While Tag@ <> TagCRTUsageQualifier

  Rem  Check the Usage Qualifier mask
  If GetFirstByte@ (CRT$) <> 1 Then SW1SW2 = swMissingSMObject : Exit
  If (GetFirstByte@ (CRT$) And UQMask) = 0 Then CheckAuthCRT = False

End Function

Rem  Function InCreationState (filenum)
Rem
Rem  Check the Life Cycle Status byte

Function InCreationState (filenum)

  Private LCS$ : Call TLVReadValue (filenum, Chr$(TagFCP)+Chr$(TagLCS), LCS$)

  Select Case Len(LCS$)

    Case 0 ' LCS not present -- do nothing

    Case 1
      Private LCS@ : LCS@ = Asc(LCS$)
      If LCS@ = 1 Then
        InCreationState = True
      ElseIf (LCS@ And &H0D) <> &H05 Then ' Operational state
        SW1SW2 = swConditionsOfUse : Exit
      End If

    Case Else
      SW1SW2 = swFileStructure : Exit

  End Select

End Function

Function CheckSCDataObject (Tag@, ByVal S$)

  Private Len@

  Select Case Tag@

    Case TagSCAlways
      CheckSCDataObject = True
      Exit Function

    Case TagSCNever
      Exit Function ' Return False

    Case TagSCByte
      CheckSCDataObject = CheckSCByte (Asc(S$))
      Exit Function

    Case TagCrtAT ' This is the only CRT type in the SCOSTA card SC's
      Private Fulfilled@
      Do
        Tag@ = GetFirstByte@ (S$)
        If GetFirstByte@ (S$) <> 1 Then_ ' Length byte must be 1
          SW1SW2 = swBadSMObject : Exit
        If Tag@ = TagCRTUsageQualifier Then Exit Do
        If Tag@ <> TagCRTSecretKey Then SW1SW2 = swBadSMObject : Exit
        Tag@ = GetFirstByte@ (S$) ' Key reference

        Rem  Set bits in Fulfilled@ as if in Usage qualifier byte
        If Tag@ = SecStat.ExtAuthRef Then Fulfilled@ = Fulfilled@ Or &H80
        If Tag@ = SecStat.UserAuthRef Or InvalidatedPassword (Tag@) Then_
          Fulfilled@ = Fulfilled@ Or &H08
      Loop
      Tag@ = GetFirstByte@ (S$) ' Usage qualifier byte
      If (Tag@ And Not Fulfilled@) = 0 Then CheckSCDataObject = True

    Case TagSCAnd
      While Len(S$) <> 0
        Tag@ = GetFirstByte@ (S$)
        Len@ = GetFirstByte@ (S$)
        If Not CheckSCDataObject (Tag@, Left$(S$, Len@)) Then
          Exit Function ' Return False
        End If
      Wend

    Case TagSCOr
      While Len(S$) <> 0
        Tag@ = GetFirstByte@ (S$)
        Len@ = GetFirstByte@ (S$)
        If CheckSCDataObject (Tag@, Left$(S$, Len@)) Then
          CheckSCDataObject = True
          Exit Function
        End If
      Wend

    Case Else
      SW1SW2 = swFunctionNotSupported : Exit

  End Select

End Function

Rem  Function InvalidatedPassword (key@)
Rem
Rem  Returns True if the password exists and is invalidated

Function InvalidatedPassword (key@)
  Call FindKey (EFIDPassword, key@)
  If Len(CurrentKey$) <> 0 Then
    If (Asc(CurrentKey$) And &H80) = 0 Then InvalidatedPassword = True
    Close CurrentKey.filenum@
  End If
End Function
