Declare Sub FindGlobalKey (EFID@, Key@)
Declare Sub FindLocalKey (EFID@, Key@)
Declare Function FindKeyInFile (filename$, Key@)

Rem  Sub FindP2Key()
Rem
Rem  Find the key referenced in P2
Rem
Rem      +---+---+---+---+---+---+---+---+
Rem  P2: | G | 0 | 0 |     Key number    |
Rem      +---+---+---+---+---+---+---+---+
Rem
Rem        G=0 => Global key (in MF)
Rem        G=1 => Local key (in current DF or some parent DF)
Rem
Rem  On exit, CurrentKey$ and CurrentKey As FileObject are set up

Sub FindP2Key (EFID@)

  If P2 And &H60 Then SW1SW2 = swP1P2Error : Exit
  Call FindKey (EFID@, P2)
  If Len(CurrentKey$) = 0 Then SW1SW2 = swDataNotFound : Exit

End Sub

Sub FindKey (EFID@, key@)

  CurrentKey$ = ""
  If key@ And &H80 Then
    Call FindLocalKey (EFID@, key@ And &H7F)
  Else
    Call FindGlobalKey (EFID@, key@)
  End If

End Sub

Rem  Sub FindGlobalKey (EFID@, Key@)
Rem
Rem  Look for a key in MF

Sub FindGlobalKey (EFID@, Key@)
  Private filename$ : filename$ = FindShortID$ (EFID@, "\MF\")
  Call FindKeyInFile (filename$, Key@)
End Sub

Rem  Sub FindLocalKey (EFID@, Key@)
Rem
Rem  Look for a key in the CurrentDF and all parent DF's up to
Rem  (but not including) MF

Sub FindLocalKey (EFID@, Key@)
  Private dir$ : dir$ = CurrentDF+"\"
  While Len(dir$) >= 9
    Private filename$ : filename$ = FindShortID$ (EFID@, dir$)
    If FindKeyInFile (filename$, Key@) Then Exit Sub
    dir$ = Left$(dir$, Len(dir$)-5) ' Parent DF
  Wend
End Sub

Rem  Function FindKeyInFile (filename$, Key@)
Rem
Rem  Look for the key in the given file. If found, sets up CurrentKey
Rem  structure and returns True.

Function FindKeyInFile (filename$, Key@)

  FindKeyInFile = False
  If Len(filename$) = 0 Then Exit Function

  CurrentKey.filenum@ = Open filename$ For Binary Access Read Write Shared
  If FileError <> 0 Then FileError = 0 : Exit Function

  Private RecLen@

  Rem  Skip the FCP
  Get #CurrentKey.filenum@, , RecLen@
  If RecLen@ <> TagFCP Then Exit Function
  Get #CurrentKey.filenum@, , RecLen@
  Seek #CurrentKey.filenum@, RecLen@+3
  If FileError <> 0 Then Exit Function

  Do
    Get #CurrentKey.filenum@, , RecLen@
    CurrentKey.filepos = Seek (#CurrentKey.filenum@)
    Get #CurrentKey.filenum@, , CurrentKey$, RecLen@
    If FileError <> 0 Then Exit Do
    If RecLen@ <> 0 And (Asc(CurrentKey$) And &H7F) = Key@ Then ' Found key
      FindKeyInFile = True
      Exit Function
    End If
  Loop

  Close CurrentKey.filenum@
  FileError = 0

End Function
