Rem  Maintain a list of (DFName, path) pairs

#IfNotDef DFNameDefined

Const DFNameBlockLen = 8 ' Allocate in blocks of size 8

Eeprom Dynamic DFName$(0 To 1, 1 To DFNameBlockLen) ' The list
Eeprom nDFNames = 0                    ' Number of DF's in list
Eeprom nDFNameEntries = DFNameBlockLen ' Size of list (multiple of DFNameBlockLen)

#EndIf

Declare Sub MoveStringPairs (Dest$(), Src$(), N)

Rem  Function FindDFName$ (name$)
Rem
Rem  Given a DF name, returns its path (or empty string if not found)

Function FindDFName$ (name$)
  Private I : For I = 1 To nDFNames
    If DFName$(0, I) = name$ Then FindDFName$ = DFName$(1, I) : Exit Function
  Next I
End Function

Rem  Sub AddDFName (name$, path$)
Rem
Rem  Add (name$, path$) to the list

Sub AddDFName (name$, path$)

  Rem  Use an empty entry if there is one
  Private I : For I = 1 To nDFNames
    If Len(DFName$(0, nDFNames)) = 0 Then
      DFName$(0, I) = name$
      DFName$(1, I) = path$
      Exit Sub
    End If
  Next I

  If nDFNames = nDFNameEntries Then

    Rem  Add DFNameBlockLen entries to the list

    Rem  - Move all the strings to a temporary location
    Private Dynamic Temp$(,)
    ReDim Temp$(0 To 1, 1 To nDFNameEntries)
    Call MoveStringPairs (Temp$, DFName$, nDFNames)

    Rem  - Reallocate DFName$
    nDFNameEntries = nDFNameEntries + DFNameBlockLen
    ReDim DFName$(0 To 1, 1 To nDFNameEntries)

    Rem  - Restore the strings to DFName$
    Call MoveStringPairs (DFName$, Temp$, nDFNames)

  End If

  nDFNames = nDFNames + 1
  DFName$(0, nDFNames) = name$
  DFName$(1, nDFNames) = path$

End Sub

Rem  Sub RemoveDFName (name$)
Rem
Rem  Just set name and path to ""

Sub RemoveDFName (name$)
  Private I : For I = 1 To nDFNames
    If DFName$(0, I) = name$ Then
      DFName$(0, I) = ""
      DFName$(1, I) = ""
      Exit Sub
    End If
  Next I
End Sub

Sub MoveStringPairs (Dest$(), Src$(), N)
  Private I : For I = 1 To N
    Dest$(0, I) = Src$(0, I) : Src$(0, I) = ""
    Dest$(1, I) = Src$(1, I) : Src$(1, I) = ""
  Next I
End Sub

Rem  Sub FindPartialDFName (prefix$)
Rem
Rem  Find the first/last/next/previous DF name that starts as specified

Sub FindPartialDFName (prefix$)

  Private Start, Stop, Direction
  If (P2 And &H01) = 0 Then ' First or next
    Start = 1
    Stop = nDFNames
    Direction = 1
  Else                      ' Last or previous
    Start = nDFNames
    Stop = 1
    Direction = -1
  End If

  Private I
  If (P2 And &H02) <> 0 Then ' Next or previous - find the current directory
    For I = Start To Stop Step Direction
      If DFName$(1, I) = CurDir Then Start = I + Direction : Exit For
    Next I
  End If

  For I = Start To Stop Step Direction
    If Left$(DFName$(0, I), Lc) = prefix$ Then
      Call SetCurrentDF (DFName$(1, I))
      Exit Sub
    End If
  Next I

  SW1SW2 = swFileNotFound : Exit

End Sub
