Sub OpenEF (FileID, Type@)

  If FileID = &HFFFF Then SW1SW2 = swNoCurrentEF : Exit

  Private filename$ : filename$ = Right$("000"+Hex$(FileID), 4)
  Private Attr% : Attr% = GetAttr (filename$)
  If FileError <> 0 Then SW1SW2 = swFileNotFound : Exit
  If (Attr% And faDirectory) <> 0 Then SW1SW2 = swNotEF : Exit

  EFFilenum = Open filename$ For Binary Access Read Write Shared
  If FileError <> 0 Then SW1SW2 = swFileNotFound : Exit

  Private tag@ : Get #EFFilenum, , tag@
  If FileError <> 0 Or tag@ <> TagFCP Then SW1SW2 = swFileStructure : Exit
  EFFCPLen = TLVReadLength (EFFilenum)
  If EFFCPLen = 0 Then SW1SW2 = swFileStructure : Exit
  EFFCPLen = EFFCPLen + Seek (#EFFilenum) - 1

  Call TLVReadValue (EFFilenum,_
    Chr$(TagFCP)+Chr$(TagFileDescriptor), EFDescriptor$)
  If EFDescriptor$ = "" Then SW1SW2 = swFileStructure : Exit

  If InvalidEFStructure (Type@) Then SW1SW2 = swFileStructure : Exit

End Sub

Function InvalidEFStructure (Type@)
  Private EFStructure@
  EFStructure@ = Asc(EFDescriptor$) And FileStructureMask
  Select Case Type@
    Case EFTypeTransparent: InvalidEFStructure = (EFStructure@ <> FSTransparent)
    Case EFTypeRecord: InvalidEFStructure = (EFStructure@ = FSTransparent)
  End Select
End Function

Function Record$ (ByVal recnum)

  Private len@ : Call FindRecord (recnum, len@)
  Get #EFFilenum, , Record$, len@
  If FileError <> 0 Then SW1SW2 = swRecordNotFound : Exit

End Function

Sub FindRecord (ByVal recnum, len@)

  Select Case Asc (EFDescriptor$) And &H06

    Case &H00 : SW1SW2 = swFileStructure : Exit

    Case &H02 ' Linear fixed

      recnum = recnum - 1 ' Zero-based

SeekFixedLength:

      len@ = RecordLength()
      Seek #EFFilenum, EFFCPLen + recnum * len@ + 1

    Case &H04 ' Linear variable
      Private pos& : pos& = EFFCPLen + 1
      Do
        Get #EFFilenum, pos&, len@
        If FileError <> 0 Then Exit Do
        If recnum = 1 Then Exit Sub
        recnum = recnum - 1
        pos& = pos& + len@ + 1
      Loop

    Case &H06 ' Cyclic
      Rem  Calculate the zero-based record index
      Private newest@ : newest@ = NewestRecord@()
      recnum = newest@ - recnum
      If recnum < 0 Then
        recnum = recnum + NumRecords()
        If recnum < newest@ Then SW1SW2 = swRecordNotFound : Exit
      End If
      GoTo SeekFixedLength

  End Select

  If FileError <> 0 Then SW1SW2 = swRecordNotFound : Exit

End Sub

Function NumRecords()

  Select Case Asc (EFDescriptor$) And &H06

    Case &H00 : SW1SW2 = swFileStructure : Exit

    Case &H02, &H06 ' Linear fixed (&H02) or Cyclic (&H06)
      NumRecords = (Len (#EFFilenum) - EFFCPLen) / RecordLength()

    Case &H04 ' Linear variable
      NumRecords = 0
      Seek #EFFilenum, EFFCPLen + 1
      Do
        If EOF(#EFFilenum) Then Exit Function
        Private Len@ : Get #EFFilenum, , Len@
        Seek #EFFilenum, Seek(#EFFilenum) + Len@
        If FileError <> 0 Then SW1SW2 = swRecordNotFound : Exit
        NumRecords = NumRecords + 1
      Loop

  End Select

End Function

Function RecordLength()

  If Len(EFDescriptor$) = 3 Then
    RecordLength = Asc(EFDescriptor$(3))
  ElseIf Len(EFDescriptor$) > 3 Then
    RecordLength = 256*Asc(EFDescriptor$(3)) + Asc(EFDescriptor$(4))
  Else
    SW1SW2 = swFileStructure : Exit
  End If

End Function

Rem  Function NewestRecord@()
Rem
Rem  Returns the (1-based) index of the newest record in a cyclic file.
Rem  This is stored under TagProprietary = &H85 in the FCP.

Function NewestRecord@()

  Private Value$ As String*1 At NewestRecord@
  Call TLVReadValue (EFFilenum, Chr$(TagFCP)+Chr$(TagProprietary), Value$)
  If FileError <> 0 Then SW1SW2 = swRecordNotFound : Exit

End Function

Function PureDataBytes()
  PureDataBytes = TLVReadIntValue (EFFilenum, Chr$(TagFCP)+Chr$(TagPureDataBytes))
End Function

Function MaxRecords()
  Select Case Len(EFDescriptor$)
    Case 5 : MaxRecords = Asc(EFDescriptor$(5))
    Case 6 : MaxRecords = 256*Asc(EFDescriptor$(5)) + Asc(EFDescriptor$(6))
    Case Else : SW1SW2 = swFileStructure : Exit
  End Select
End Function

Rem  Sub OpenBinaryEF()
Rem
Rem  Common code for READ BINARY / WRITE BINARY / UPDATE BINARY

Sub OpenBinaryEF()

  If P1 And &H80 Then
    Call SelectByShortID (P1 And &H7F, EFTypeTransparent)
    P1 = 0
  Else
    Call OpenEF (CurrentEF, EFTypeTransparent)
  End If

  If (DataCodingByte() And DCWriteMask) = DCWriteOnce Then
    WrittenMaskLen = (PureDataBytes() + 7) / 8
  Else
    WrittenMaskLen = 0
  End If

End Sub

Rem  Function ReadBinaryBytes (ByVal Offset, ByVal Len@)
Rem
Rem  Read bytes from a transparent file, taking erased state into account

Function ReadBinaryBytes$ (ByVal Offset, ByVal Len@)

  If Offset + Len@ > PureDataBytes() Then Len@ = PureDataBytes() - Offset

  Offset = Offset + EFFCPLen + WrittenMaskLen
  Private N : N = Len(#EFFilenum) - Offset
  If N > Len@ Then N = Len@
  If N < 0 Then
    N = 0
    ReadBinaryBytes$ = ""
  Else
    Get #EFFilenum, Offset+1, ReadBinaryBytes$, N
    If FileError <> 0 Then SW1SW2 = swExecutionError : Exit
  End If
  ReadBinaryBytes$ = ReadBinaryBytes$ + String$(Len@-N, ErasedState@())

End Function

Rem  Function DataCodingByte()
Rem
Rem  Search for the data coding byte, returning default value if not found

Function DataCodingByte()

  DataCodingByte = &H41 ' Default (Write-OR, data unit size = 2 quartets)

  If Len(EFDescriptor$) >= 2 Then_
    DataCodingByte = Asc(EFDescriptor$(2)) : Exit Function

  Rem  Look in successive parent directories

  Private dir$ : dir$ = CurDir
  Do
    Private filenum
    filenum = Open dir$+"\FCP" For Binary Access Read
    If FileError = 0 Then
      Private FD$
      Call TLVReadValue (filenum, Chr$(TagFCP)+Chr$(TagFileDescriptor), FD$)
      Close filenum
      If Len(FD$) >= 2 Then_
        DataCodingByte = Asc(FD$(2)) : Exit Function
    End If
    FileError = 0
    If Len(dir$) <= 5 Then Exit Function
    dir$ = Left$(dir$, Len(dir$)-5) ' Parent directory
  Loop

End Function

Function ErasedState@()
  Private DCB@ : DCB@ = DataCodingByte()
  ErasedState@ = &H00
  If (DCB@ And DCWriteMask) = DCWriteAND Then ErasedState@ = &HFF
End Function
