Rem  EXTERNAL AUTHENTICATE
Rem  MUTUAL AUTHENTICATE
Rem
Rem  These two commands have the same INS value, and are
Rem  distinguished only by whether Le is present or not
Rem
Rem  ISO/IEC 7816-4 section 7.5.3
Rem  SCOSTA sections 6.2.3, 6.2.4

Sub ExternalAuthenticate (IDATA$)

  If P1 > 1 Then SW1SW2 = swConditionsOfUse : Exit

  If LePresent() Then ' MUTUAL AUTHENTICATE
    If Lc <> 16 Then SW1SW2 = swLcLeError : Exit
  Else                ' EXTERNAL AUTHENTICATE
    If Lc <> 0 And Lc <> 8 Then SW1SW2 = swLcLeError : Exit
  End If

  If ChallengeState <> ChallengeAvailable Then SW1SW2 = swDataNotFound : Exit

  Call CheckAccess (0, 0)

  Rem  Find the key
  Call FindP2Key (EFIDKey)

  Rem  Check that it's valid
  If (Asc(CurrentKey$) And &H80) = 0 Then ' V bit clear (key invalid)
    If Lc <> 0 Then SW1SW2 = swReferenceDataInvalidated
    Exit
  End If

  Rem  Check that it can be used for External Authentication
  Private KeyType@ : KeyType@ = Asc(CurrentKey$(2))
  If (KeyType@ And &H01) = 0 Then SW1SW2 = swConditionsOfUse : Exit

  Rem  Get the index of the Retry Counter
  Private I = 3
  If KeyType@ And &H20 Then I = I + 2
  If KeyType@ And &H10 Then I = I + 2
  If KeyType@ And &H02 Then I = I + 2

  Rem  Check the retry counter in the high nibble of CurrentKey$(I)
  Private RetryCounter@ : RetryCounter@ = Asc(CurrentKey$(I))
  If RetryCounter@ < &H10 Then SW1SW2 = swKeyDisabled : Exit

  If Lc = 0 Then ' Just return the retry count
    SW1SW2 = swRetriesRemaining + RetryCounter@ / 16
    Exit
  End If

  Rem  Check that the Algorithm Reference is 3DES
  If Asc(CurrentKey$(I+1)) > 1 Then SW1SW2 = swConditionsOfUse : Exit

  CurrentKey$ = Mid$(CurrentKey$, I+2)

  If RetryCounter@ < &HF0 Then
    Rem  Decrement the retry counter before checking IDATA$,
    Rem  to prevent current-based attacks
    RetryCounter@ = RetryCounter@ - &H10
    Put #CurrentKey.filenum@, CurrentKey.filepos+I-1, RetryCounter@
    If FileError <> 0 Then SW1SW2 = swFileWriteError : Exit
  End If

  Rem  Check IDATA$
  Call Encrypt3DESCBC (ChallengeData$, CurrentKey$)
  If Left$(IDATA$, 8) = ChallengeData$ Then ' Authentication successful
    If RetryCounter@ < &HF0 Then
      Rem  Reset the Retry Counter
      RetryCounter@ = &H11 * (RetryCounter@ And &H0F)
      Put #CurrentKey.filenum@, CurrentKey.filepos+I-1, RetryCounter@
      If FileError <> 0 Then SW1SW2 = swFileWriteError : Exit
    End If

    If LePresent() Then ' MUTUAL AUTHENTICATE
      IDATA$ = Right$(IDATA$, 8)
      Call InternalAuthenticate (IDATA$)
    End If

    SecStat.ExtAuthRef = P2 ' The result of all this

  Else ' Authentication failed
    SW1SW2 = swRetriesRemaining + RetryCounter@ / 16
  End If

End Sub
