Rem  SCOSTA.Bas
Rem
Rem  Secure Card Operating System for Transport Applications
Rem
Rem  Main module

Option Explicit
Option Base Explicit

Public CurrentDF As String  ' Filename
Public CurrentEF            ' File ID
Public EFFilenum            ' Filenum of Current EF if open
Public EFDescriptor$
Public EFFCPLen
Public WrittenMaskLen       ' for write-once binary files
Public RecordPointer As Byte

#Include FileIO.Def
#Include Misc.Def

#Include Types.Def
#Include SecStat.Def
#Include TLV.Def
#Include EF.Def
#Include SM.Def
#Include Commands.Def
#Include FileIO.Def
#Include SW1SW2.Def

#Include Externs.Def

#Include LePresnt.Bas
#Include DFNames.Bas
#Include TLV.Bas
#Include EF.Bas
#Include ShortID.Bas
#Include Access.Bas
#Include AlgoKey.Bas
#Include DES.Bas
#Include SetLCS.Bas
#Include Utils.Bas

Rem  SCOSTA 6.1: File related commands in SCOSTA

#Include ReadBin.Bas
#Include ReadRec.Bas
#Include WriteBin.Bas
#Include WriteRec.Bas
#Include UpdBin.Bas
#Include UpdRec.Bas
#Include AppRec.Bas
#Include EraseBin.Bas
#Include Create.Bas
#Include Delete.Bas
#Include Select.Bas
#Include Activate.Bas
#Include Deact.Bas
#Include TermDF.Bas
#Include TermEF.Bas
#Include TermCard.Bas

Rem  SCOSTA 6.2: Security related commands in SCOSTA

#Include Verify.Bas
#Include IntAuth.Bas
#Include ExtAuth.Bas ' includes MUTUAL AUTHENTICATE
#Include GetChall.Bas
#Include MSE.Bas
#Include PSO.Bas
#Include EnabVR.Bas
#Include DisabVR.Bas
#Include ChangeRD.Bas
#Include ResetRC.Bas

Rem  SCOSTA 6.3: Other Commands

#Include GetData.Bas
#Include PutData.Bas

Rem  GET RESPONSE is handled by the operating system T=0 code

Rem  Initialisation code

CurrentDF = "\MF" ' ISO Master File (assumed to exist)
CurrentEF = &HFFFF

SecStat.UserAuthRef = &HFF
SecStat.ExtAuthRef = &HFF

Rem  Command Dispatcher

Command Else CommandDispatcher (IDATA$)

  If CardTerminated@ Then SW1SW2 = swConditionsOfUse : Exit

  Close ' Close all files
  EFFilenum = 0
  FileError = 0

  If ChallengeState <> 0 Then ChallengeState = ChallengeState - 1

  Rem  The Channel number in the bottom two bits of CLA must be zero:

  If (CLA And &H03) <> 0 Then SW1SW2 = swChannelNotSupported : Exit

  ChDir CurrentDF
  If FileError <> 0 Then
    FileError = 0
    CurrentDF = "\MF"
    ChDir CurrentDF
  End If

  SMCoding@ = CLA And &H0C

  Select Case INS

    Rem  SCOSTA 6.1: File related commands in SCOSTA

    Case &HB0 : Call ReadBinary (IDATA$)
    Case &HB2 : Call ReadRecords (IDATA$)
    Case &HD0 : Call WriteBinary (IDATA$)
    Case &HD2 : Call WriteRecord (IDATA$)
    Case &HD6 : Call UpdateBinary (IDATA$)
    Case &HDC : Call UpdateRecord (IDATA$)
    Case &HE2 : Call AppendRecord (IDATA$)
    Case &H0E : Call EraseBinary (IDATA$)
    Case &HE0 : Call CreateFile (IDATA$)
    Case &HE4 : Call DeleteFile (IDATA$)
    Case &HA4 : Call SelectFile (IDATA$)
    Case &H44 : Call ActivateFile (IDATA$)
    Case &H04 : Call DeactivateFile (IDATA$)
    Case &HE6 : Call TerminateDF (IDATA$)
    Case &HE8 : Call TerminateEF (IDATA$)
    Case &HFE : Call TerminateCardUsage (IDATA$)

    Rem  SCOSTA 6.2: Security related commands in SCOSTA

    Case &H20 : Call Verify (IDATA$)
    Case &H88 : Call InternalAuthenticate (IDATA$)
    Case &H82 : Call ExternalAuthenticate (IDATA$)
    Case &H84 : Call GetChallenge (IDATA$)
    Case &H22 : Call ManageSecurityEnvironment (IDATA$)
    Case &H2A : Call PerformSecurityOperation (IDATA$)
    Case &H28 : Call EnableVerificationRequirement (IDATA$)
    Case &H26 : Call DisableVerificationRequirement (IDATA$)
    Case &H24 : Call ChangeReferenceData (IDATA$)
    Case &H2C : Call ResetRetryCounter (IDATA$)

    Rem  SCOSTA 6.3: Other Commands

    Case &HCA : Call GetData (IDATA$)
    Case &HDA : Call PutData (IDATA$)

  End Select

End Command