Rem  SELECT FILE
Rem
Rem  ISO/IEC 7816-4 section 7.1.1
Rem  SCOSTA section 6.1.11

Sub SelectFile (S$)

  Call FindSelectedFile (S$)
  Call BuildSelectResponse (S$)

End Sub

Sub FindSelectedFile (S$)

  Select Case P1

    Case 0 ' Select by file ID

      If Lc = 0 Then
        Call SetCurrentDF ("\MF")
      Else
        If Lc <> 2 Then SW1SW2 = swLcInconsistent : Exit

        S$ = BinToHex$ (S$)
        If S$ = "3F00" Then
          Call SetCurrentDF ("\MF")
        Else

          Rem  FileID must be unique in:
          Rem    - the immediate children of the current DF
          Rem    - the parent DF
          Rem    - the immediate children of the parent DF

          Private nMatches = 0
          Private filename$, found$
          filename$ = S$
          If Dir(filename$) = 1 Then ' Child of current DF matched
            found$ = CurDir+"\"+filename$
            nMatches = 1
          End If
          filename$ = CurDir
          If Len(filename$) > 3 Then
            filename$ = Left$(filename$, Len(filename$)-5)

            If Right$(filename$, 4) = S$ Then ' Parent DF matched
              found$ = filename$
              nMatches = nMatches + 1
            End If
            filename$ = filename$+"\"+S$
            If Dir(filename$) = 1 Then ' Child of parent DF matched
              found$ = filename$
              nMatches = nMatches + 1
            End If
          End If

          If nMatches <> 1 Then SW1SW2 = swFileNotFound : Exit
          Call SetEFDF (found$)

        End If

      End If

    Case 1 ' Select child DF

      If Lc <> 2 Then SW1SW2 = swLcInconsistent : Exit
      Call SetCurrentDF (CurDir + "\" + BinToHex$ (S$))

    Case 2 ' Select child EF

      If Lc <> 2 Then SW1SW2 = swLcInconsistent : Exit
      Call SetCurrentEF (CurDir + "\" + BinToHex$ (S$), EFTypeAny)

    Case 3 ' Select parent DF

      If S$ <> "" Then SW1SW2 = swLcInconsistent : Exit

      If CurDir = "\MF" Then SW1SW2 = swFileNotFound : Exit

      Rem  Shorten the CurrentDF name by 5 bytes
      Private Len% : Len% = Len (CurDir)
      Call SetCurrentDF (Left$ (CurDir, Len%-5))

    Case 4 ' Select DF by name

      Call FindPartialDFName (S$)

    Case 8 ' Select from MF

      Call SetEFDF ("\MF\" + BinToHexPath$ (S$))

    Case 9 ' Select from current DF

      Call SetEFDF (CurDir + "\" + BinToHexPath$ (S$))

  End Select

End Sub

Sub BuildSelectResponse (S$)

  If P2 And &HF0 Then SW1SW2 = swP1P2Error : Exit
  If Not LePresent() Then Exit Sub

  P2 = P2 And &H0C
  Select Case P2 And &H0C
    Case &H08 : S$ = Chr$(TagFMD) + Chr$(0) : Exit Sub ' Empty FMD template
    Case &H0C : S$ = "" : Exit Sub
  End Select

  Private filenum, length
  If CurrentEF = &HFFFF Then ' DF
    filenum = Open "FCP" For Binary Access Read Shared
    length = Len(#filenum)
  Else
    filenum = EFFilenum
    Seek #filenum, 1
    length = EFFCPLen
  End If

  If length > 255 Then length = 255
  If Le <> 0 And length > Le Then length = Le

  Get #filenum, , S$, length
  If FileError <> 0 Then SW1SW2 = swDataNotFound : Exit

  If (P2 And &H0C) = 0 Then ' Wrap it in an FCI template
    S$ = Chr$(TagFCI) + Chr$(Len(S$)) + S$
  End If

End Sub

Sub SetCurrentDF (S$)

  Private Attr% : Attr% = GetAttr (S$)
  If FileError <> 0 Then SW1SW2 = swFileNotFound : Exit
  If (Attr% And faDirectory) = 0 Then SW1SW2 = swNotDF : Exit

  CurrentDF = S$
  ChDir S$

  CurrentEF = &HFFFF
  RecordPointer = 0
  SecStat.UserAuthRef = &HFF
  SecStat.ExtAuthRef = &HFF

End Sub

Sub SetCurrentEF (S$, Type@)

  Private Attr% : Attr% = GetAttr (S$)
  If FileError <> 0 Then SW1SW2 = swFileNotFound : Exit
  If (Attr% And faDirectory) <> 0 Then SW1SW2 = swNotEF : Exit

  Private fn : fn = Open S$ For Binary Access Read Write Shared
  If FileError <> 0 Then SW1SW2 = swFileNotFound : Exit
  Private tag@ : Get #fn, , tag@
  If FileError <> 0 Or tag@ <> TagFCP Then SW1SW2 = swFileStructure : Exit
  EFFCPLen = TLVReadLength (fn)
  If EFFCPLen = 0 Then SW1SW2 = swFileStructure : Exit
  EFFCPLen = EFFCPLen + Seek (#fn) - 1

  Call TLVReadValue (fn, Chr$(TagFCP)+Chr$(TagFileDescriptor), EFDescriptor$)
  If EFDescriptor$ = "" Then SW1SW2 = swFileStructure : Exit

  If InvalidEFStructure (Type@) Then SW1SW2 = swFileStructure : Exit

  EFFilenum = fn

  Rem  We want to say
  Rem
  Rem     CurrentEF = ValH (Right$(S$, 4))
  Rem
  Rem  But this overflows for values >= &H8000. So:

  Private ValH& : ValH& = ValH (Right$(S$, 4))
  Private ValH% At ValH& + 2

  Private ParentDir$ : ParentDir$ = Left$(S$, Len(S$) - 5)
  If ParentDir$ <> CurDir Then

    Rem  Set CurrentDF to parent DF of selected EF
    Call SetCurrentDF (ParentDir$)

  Else

    Rem  Reset RecordPointer if CurrentEF has changed
    If CurrentEF <> ValH% Then RecordPointer = 0

  End If

  CurrentEF = ValH%

End Sub

Rem  Sub SetEFDF (S$)
Rem
Rem  Set CurrentDF or CurrentEF, according to the file type

Sub SetEFDF (S$)
  If (GetAttr (S$) And faDirectory) <> 0 Then
    Call SetCurrentDF (S$)
  Else
    Call SetCurrentEF (S$, EFTypeAny)
  End If
End Sub

Rem  Function FilenameFromPath$ (S$)
Rem
Rem  On input, S$ contains an ISO path, i.e. a sequence of 2-byte file
Rem  identifiers.
Rem
Rem  This function constructs a BasicCard path from S$, i.e. a sequence
Rem  of 4-byte ASCII hex filenames separated internally by backslashes

Function FilenameFromPath$ (ByVal S$)
  S$ = BinToHexPath$ (S$)
  Private Prefix$ As String*4 : Prefix$ = S$
  S$ = Mid$(S$, 5)
  If Prefix$ = "3F00" Then
    FilenameFromPath$ = "\MF" + S$
  Else
    If Prefix$ <> "3FFF" Then
      Rem  Prefix$ must be the current DF
      If Prefix$ <> Right$(CurDir, 4) Then SW1SW2 = swFileNotFound : Exit
    End If
    FilenameFromPath$ = CurDir + S$
  End If
End Function

Rem  Function BinToHexPath$ (S$)
Rem
Rem  Convert a binary path to an ASCII hex path with separators

Function BinToHexPath$ (ByVal S$)
  Do
    If S$ = "" Then Exit Do
    If BinToHexPath$ <> "" Then BinToHexPath$ = BinToHexPath$ + "\"
    BinToHexPath$ = BinToHexPath$ + BinToHex$ (S$)
    S$ = Mid$ (S$, 3)
  Loop
End Function

Rem  Function BinToHex$ (ByVal S$ As String*2)
Rem
Rem  Convert a 2-byte binary string to a 4-byte ASCII hex string

Function BinToHex$ (ByVal S$ As String*2)
  Private N At S$
  BinToHex$ = Right$ ("000"+Hex$(N), 4)
End Function
