Rem  WRITE BINARY
Rem
Rem  ISO/IEC 7816-4 section 7.2.2

Sub WriteBinary (S$)

  If LePresent() Then SW1SW2 = swLcLeError : Exit

  Call OpenBinaryEF()

  Call CheckAccess (&H04, 0) ' &H04 from Table 18
  
  If P1P2 + Lc > PureDataBytes() Then SW1SW2 = swOutOfMemory : Exit

  Private DCB@ : DCB@ = DataCodingByte() And DCWriteMask
  If DCB@ = DCWriteAND Or DCB@ = DCWriteOR Then
    Rem  Read the relevant bytes from the file
    Private Old$ : Old$ = ReadBinaryBytes$ (P1P2, Lc)
  End If

  Private I

  Select Case DCB@

    Case DCWriteOnce
      If GetWrittenState (P1P2, Lc) <>  WSErased Then_
        SW1SW2 = swConditionsOfUse : Exit ' ???
      Call SetWrittenState (P1P2, Lc, WSWritten)

    Case DCWriteAND
      For I = 1 To Lc
        S$(I) = Chr$(Asc(S$(I)) And Asc(Old$(I)))
      Next I

    Case DCWriteOR
      For I = 1 To Lc
        S$(I) = Chr$(Asc(S$(I)) Or Asc(Old$(I)))
      Next I

  End Select

  Put #EFFilenum, P1P2+EFFCPLen+WrittenMaskLen+1, S$
  If FileError <> 0 Then SW1SW2 = swFileWriteError : Exit

End Sub

Rem  Procedures to handle the WrittenMask after the FCP. This is a
Rem  bit mask of all bytes that have been written in a write-once file
Rem
Rem  Function GetWrittenState (Start, Length)
Rem
Rem    Returns the state of the byte range: WSErased, WSWritten, or WSMixed
Rem
Rem  Sub SetWrittenState (Start, Length, State@)
Rem
Rem    State@ should be WSErased or WSWritten

Type WSParams
  StartOffset, StartMask@
  EndOffset, EndMask@
End Type

Declare Sub WSSetupParams (ByVal Start, ByVal Length, WSP As WSParams)

Function GetWrittenState (Start, Length)
  Private WSP As WSParams
  Call WSSetupParams (Start, Length, WSP)

  Private Ch@

  Rem  Start byte
  Get #EFFilenum, WSP.StartOffset, Ch@
  Ch@ = Ch@ And WSP.StartMask@
  If Ch@ = WSP.StartMask@ Then
    Ch@ = &HFF
    GetWrittenState = WSWritten
  ElseIf Ch@ = 0 Then
    GetWrittenState = WSErased
  Else
    GetWrittenState = WSMixed
    Exit Function
  End If

  If WSP.StartOffset = WSP.EndOffset Then Exit Function

  WSP.StartOffset = WSP.StartOffset + 1

  Rem  Intermediate bytes, in blocks of up to 254
  While WSP.StartOffset < WSP.EndOffset
    Private N : N = WSP.EndOffset - WSP.StartOffset
    If N > 254 Then N = 254
    Private S$ : Get #EFFilenum, , S$, N
    If S$ <> String$(N, Ch@) Then GetWrittenState = WSMixed : Exit Function
    WSP.StartOffset = WSP.StartOffset + N
  Wend

  Rem  End byte
  If WSP.EndMask@ <> 0 Then
    Get #EFFilenum, , Ch@
    Ch@ = Ch@ And WSP.EndMask@
    If (Ch@ <> 0 Or GetWrittenState <> WSErased) And_
      (Ch@ <> WSP.EndMask@ Or GetWrittenState <> WSWritten) Then_
        GetWrittenState = WSMixed
  End If

  If FileError <> 0 Then SW1SW2 = swExecutionError : Exit

End Function

Sub SetWrittenState (Start, Length, State@)
  Private WSP As WSParams
  Call WSSetupParams (Start, Length, WSP)

  Private Ch@

  Rem  Start byte
  Get #EFFilenum, WSP.StartOffset, Ch@
  If State@ = WSWritten Then
    Ch@ = Ch@ Or WSP.StartMask@
  Else
    Ch@ = Ch@ And Not WSP.StartMask@
  End If
  Put #EFFilenum, WSP.StartOffset, Ch@

  If WSP.StartOffset = WSP.EndOffset Then Exit Sub

  WSP.StartOffset = WSP.StartOffset + 1

  Rem  Intermediate bytes, in blocks of up to 254
  While WSP.StartOffset < WSP.EndOffset
    Private N : N = WSP.EndOffset - WSP.StartOffset
    If N > 254 Then N = 254
    If State@ = WSWritten Then Ch@ = &HFF : Else Ch@ = 0
    Private S$ : S$ = String$(N, Ch@)
    Put #EFFilenum, , S$
    WSP.StartOffset = WSP.StartOffset + N
  Wend

  Rem  End byte
  If WSP.EndMask@ <> 0 Then
    Get #EFFilenum, , Ch@
    If State@ = WSWritten Then
      Ch@ = Ch@ Or WSP.EndMask@
    Else
      Ch@ = Ch@ And Not WSP.EndMask@
    End If
    Put #EFFilenum, WSP.EndOffset, Ch@
  End If

  If FileError <> 0 Then SW1SW2 = swExecutionError : Exit

End Sub

Sub WSSetupParams (ByVal Start, ByVal Length, WSP As WSParams)

  If Length = 0 Then
    WSP.StartOffset = 1
    WSP.EndOffset = 1
    WSP.StartMask@ = 0
    WSP.EndMask@ = 0
    Exit Sub
  End If

  Length = Start + Length - 1 ' Now Length is offset of last byte in range

  WSP.StartOffset = EFFCPLen + Start / 8 + 1
  Start = Start And 7
  WSP.StartMask@ = &HFF
  While Start
    WSP.StartMask@ = 2 * (WSP.StartMask@ And &H7F) ' And &H7F to avoid overflow
    Start = Start - 1
  Wend

  WSP.EndOffset = EFFCPLen + Length / 8 + 1
  Length = Length And 7
  WSP.EndMask@ = &H01
  While Length
    WSP.EndMask@ = 2 * WSP.EndMask@ + 1
    Length = Length - 1
  Wend

  If WSP.EndOffset = WSP.StartOffset Then
    WSP.StartMask@ = WSP.StartMask@ And WSP.EndMask@
    WSP.EndMask@ = 0
  End If

  If WSP.EndMask@ = &HFF Then
    WSP.EndOffset = WSP.EndOffset + 1
    WSP.EndMask@ = 0
  End If

End Sub